#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json

class ParameterValue(object):
    """Base class for all analysis templates parameters values

    :param str id: The unique identifier for this parameter.
    :param str value: The value of this parameter.
    """

    def __init__(self, id, value, _type):
        """ ParameterValue init """
        self.id = id
        self.value = value
        self._type = _type

    def _to_json(self, body):
        body['id'] = self.id
        body['value'] = self.value
        body['type'] = self._type

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def _from_json(self, body, execution=None):
        self.id = body['id']
        self.value = body['value']
        self._type = body['type']



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import inspect
import os
from zipfile import ZipFile, is_zipfile
import types
import importlib

__all__ = [ 'module_path', 'open_module_file' ]

def module_path(module=None):
    if isinstance(module, basestring):
        module = importlib.import_module(module)

    if module is not None:
        module_filepath = module.__file__
    else:
        # Go find the module that our caller resides in.
        # The current module would not be useful.
        # Recurse in case we're called by another method in this file.
        frame = inspect.currentframe()
        base_module_filepath = inspect.getfile(frame)
        module_filepath = base_module_filepath
        while frame and module_filepath == base_module_filepath:
            frame = frame.f_back
            module_filepath = inspect.getfile(frame)

    return os.path.dirname(
        os.path.abspath(
            module_filepath))

def open_module_file(relative_filename, module=None):
    zip_file_path, path = _zipfile_path(
        os.path.join(module_path(module), relative_filename))
    if zip_file_path:
        zip_file = ZipFile(zip_file_path)
        # os.path.join will generate a path using '\\' as the separator on Windows but the path in the zipfile needs to use '/'
        zip_contained_file = zip_file.open(path.replace("\\","/"))

        # Monkey-patch in some logic to close the containing ZipFile when
        # we're done, so we don't leak file handles.
        # Icky but effective.
        existing_close = zip_contained_file.close
        def new_close(self):
            existing_close()
            zip_file.close()
        zip_contained_file.close = types.MethodType(new_close, zip_contained_file)

        return zip_contained_file
    else:
        return open(path)

def _zipfile_path(path):
    """ Given a path that may reside inside a zipfile,
    return a tuple of either (path, '') or
    (path to containing zipfile, path within zipfile)
    as strings."""
    ## Walk the provided path from leaf to root
    if not path or not isinstance(path, basestring):
        # What did you give us?
        # Also a base case for "os.path.split()", though we should
        # never get here since the previous base case should be
        # "/", which should match a subsequent clause.
        raise IOError, "Invalid argument.  Please provide a valid filesystem path."
    elif is_zipfile(path):
        # If we found a zipfile in our path, then we are in it.
        return (path, '')
    elif os.path.exists(path):
        # If we hit a valid POSIX path
        # before we hit a zipfile,
        # then the rest of the hierarchy must be all directories.
        # No zipfile here.  Short-circuit.
        return ('', path)
    else:
        # We're probably walking up a directory tree inside a zipfile.
        # It's expensive to validate this, so don't do so here.
        # Recurse.
        dirname, basename = os.path.split(path)
        zipfile, partial_path = _zipfile_path(dirname)
        return (zipfile, os.path.join(partial_path, basename))



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from analysis import Analysis
from analysis_templates import AnalysisTemplate
from data_source import DataSource
from organization import _Organization
from search import Search
from search_templates import SearchTemplates
from session import _Session, Http404Exception
from solution import Solution
from error_metric import ErrorMetrics

import os.path
from datetime import datetime

class Eureqa:
    """Represents an interface to the Eureqa server. All interactions with the server should start from this class.

    :var ~eureqa.search_templates.SearchTemplates `~eureqa.Eureqa.search_templates`: Provides access to predefined search templates.

    :param str url: The URL of the eureqa server. It should be the same URL as used to access the web UI.
    :param str user_name:
        The user name to login into the server. It should be the same user name as used to login into the web UI.
        If the user name is not provided and the interactive mode is enabled, the user name will be requested
        during the script execution.
    :param str password:
        The password.
        If the password is not provided and the interactive mode is enabled, the password will be requested
        during the script execution.
    :param str organization:
        The name of the organization.
        All request to API will be executed in the context of this organization.
        If the organization name is not provided and the user is assigned to only one organization, then that
        organization will be used by default.
    :param bool interactive_mode:
        If set to True, enables interactive mode. In the interactive mode the script will request user name, password,
        and potentially other information if it is not provided or incorrect.
        If set to False (default), throws an exception if a login or password is incorrect, or if the two factor
        authentication is enabled. This is the default behaviour which prevents scripts from indefinitely waiting
        for the user input if they are executed automatically by a CRON job or a Windows Scheduler task.
    :param bool save_credentials:
        If set to True, saves user name and password to the '.eureqa_passwd' in the user directory. If after that any
        script on the same machine is trying to connect to the same server and does not provide credentials, it reuses
        the saved credentials.
        It does not save a temporary password when the two-factor authentication is enabled.
        When used with the two-factor authentication, the interactive_mode parameter should also be enabled.
    :param bool verify_ssl_certificate:
        If set to False will not verify SSL certificate authenticity while connecting to Eureqa.
    :param bool verify_version:
        If set to False will allow to use Python API library with incompatible version of Eureqa.
        Should only be used for the diagnostic purpose.
    :param bool verbose:
        If set to True will print to the console the detailed information for each request to the server.
        Should only be used for the diagnostic purpose.
    :param int retries:  The number of attempts to establish a session before an Exception is raised
    :param str session_key: The session identifier.
    :param int timeout_seconds:
        The HTTP connection and transmission timeout. Any request to Eureqa API will abort with an exception if it
        takes more time, than set in the timeout, to either connect to the server or to receive a next data package.
    :param str key:
        Authentication key.  Provide either this field or `password`.

    :raise Exception: If the authentication fails or cannot be completed.
    """

    # note that if session_key is present, we don't actually try to do
    # a login and instead use the existing session key directly

    from api_version import API_VERSION

    def __init__(self, url='https://rds.nutonian.com', user_name=None, password=None,
                 organization=None,
                 interactive_mode=False, save_credentials=False,
                 verify_ssl_certificate=True, verify_version=True, verbose=False, retries=5,
                 session_key=None, timeout_seconds=None, key=None):
        self._session = _Session(url, user_name, password, key, verbose, save_credentials, organization,
                                verify_ssl_certificate, retries, timeout_seconds, session_key, interactive_mode)
        if verify_version:
            self._verify_version()

    def create_data_source(self, name, file_or_path, series_id_column_name=None, series_order_column_name=None, _hidden=False):
        """Creates a new data source on the server.

        Uploads raw data and and creates a new datasource.

        :param str name: A name for the new data source.
        :param str file_or_path:
            A path to a local CSV file with data for the data source.
            It can be either an absolute path or path relative to the current working directory.
            Alternatively, a Python file-like object.
        :param str series_id_column_name: The name of the column that splits data into series based on its values.
        :param str series_order_column_name: The name of the column that defines the order of the data. This indicates the data is timeseries data and will sort the rows based on this column. Use DataSource.EXISTING_ROW_ORDER to use the current row number as the series order

        :return: A :py:class:`~eureqa.data_source.DataSource` object that represents a newly created data source on the server.
        :rtype: DataSource

        """

        # Trying to open before creating data source. So if there is a problem
        # with opening this file, a data source will not be even created.
        # Otherwise if there is a problem later, the call will terminal with exception
        # and an empty data source will stay in the system.

        if isinstance(file_or_path, basestring):
            f = open(file_or_path, 'rb')
            f.close()

        data_source = DataSource(self, datasource_name=name, hidden=_hidden)
        data_source._series_id_column = series_id_column_name
        data_source._series_order_column = series_order_column_name
        data_source._create(file_or_path)

        return data_source

    def get_all_data_sources(self):
        """Get all data sources from the server

        :return: A list of :py:class:`~eureqa.data_source.DataSource` objects for all data sources within the organization.
        :rtype: list of :py:class:`~eureqa.data_source.DataSource`
        """
        return DataSource._get_all(self)

    def get_data_source(self, data_source_name):
        """Get a data source by its name.

        Searches on the server for a data source given its name.

        :param str data_source_name: The name of the data source.
        :return: A :py:class:`~eureqa.data_source.DataSource` object if such data source exists, otherwise None.
        :rtype: DataSource
        """

        matching_sources = [ds for ds in self.get_all_data_sources() if ds.name == data_source_name]
        if len(matching_sources) == 0:
            message = 'Unknown datasource_name: \'%s\'' % data_source_name
            raise Exception(message)
        elif len(matching_sources) == 1:
            return matching_sources[0]
        else:
            ids = ''
            for data_source in matching_sources:
                ids += data_source._data_source_id + " "
            message = 'Multiple data sources exist with the name: \'%s\' Use get_data_source_by_id instead with one of the following ids: %s' %(data_source_name, ids)
            raise Exception(message)

    def get_data_source_by_id(self, data_source_id):
        """Get a data source by its id.

        Searches on the server for a data source given its id.

        :param str data_source_id: The ID of the data source.
        :return: A :py:class:`~eureqa.data_source.DataSource` object if such data source exists, otherwise None.
        :rtype: DataSource
        """
        return DataSource._get(self, datasource_id = data_source_id)

    def create_analysis(self, name, description=None):
        """Creates an analysis.

        :param str name: The analysis name. It will be used as the Analysis title.
        :param str description: The analysis description.
        :return: An :py:class:`~eureqa.analysis.Analysis` object that represents a newly created analysis on the server.
        :rtype: Analysis
        """

        body = {'analysis_name': name, 'description': description, 'items': []}
        self._session.report_progress('Creating analysis: \'%s\'.' % name)
        result = self._session.execute('/analysis', 'POST', body)
        return Analysis(result, self)

    def get_analyses(self):
        """Return the list of all analyses from the server.

        :rtype: list of :py:class:`~eureqa.analysis.Analysis`
        """
        self._session.report_progress('Getting details for analyses.')
        results = self._session.execute('/analysis', 'GET')
        return [Analysis(x, self) for x in results]

    def get_analysis(self, analysis_id):
        """Return a specific analysis from the server, by id

        :param str analysis_id: The id of the analysis to return
        :rtype: Analysis
        """
        self._session.report_progress('Getting details for analysis: \'%s\'.' % analysis_id)
        results = self._session.execute('/analysis/%s' % analysis_id, 'GET')
        return Analysis(results, self)

    def create_analysis_template(self, name, description, parameters, icon=None, icon_filepath=None, _sandboxed=None):
        """ Create a new Analysis Template on the Eureqa server.

        :param str name: The analysis template's name.  Will be used to identify the template.
        :param str description: The analysis template's description.  Will be used where more space is available for an expanded description.
        :param ~eureqa.analysis_templates.Parameters parameters: Object describing the parameters that a user must fill in via the UI in order to specify the template's behavior
        :param str icon: The url of an icon to use in the UI for this analysis.
        :param str icon_filepath: The local path to an icon to upload for use in the UI for this analysis. Overrides icon argument.

        :return: An :py:class:`~eureqa.analysis_templates.AnalysisTemplate` object representing the template on the server
        :rtype: ~eureqa.analysis_templates.AnalysisTemplate """
        # :param bool _sandboxed: Should the template be run locally on the app server or remotely on a Docker execution node?
        # Deliberately undocumented; for Nutonian internal use only.
        # Docker execution nodes are not yet available in all configurations; jobs instructed to run on one will fail if one is not available.

        if not self._session._is_root():
            raise Exception("Analysis templates can only be modified by a root user. You are '%s'" %(self._session.user))

        icon_url = None
        icon_url_fill_custom = False
        if icon_filepath is not None:
            if not os.path.isfile(icon_filepath):
                raise Exception('Analysis template icon file located at "' + icon_filepath + '" does not exist')
            icon_url_fill_custom = True
        elif icon is not None:
            icon_url = icon

        body = {"name": name,
                "description": description,
                "icon_url": icon_url,
                "icon_url_fill_custom": icon_url_fill_custom}
        if _sandboxed is not None:
            body["sandboxed"] = sandboxed
        body.update(parameters._to_json())
        self._session.report_progress('Creating analysis_template: %s' % name)
        result = self._session.execute('/analysis_templates', 'POST', body)
        template = AnalysisTemplate(result, self)

        # now that we have an ID for this analysis template, upload icon file if one was provided
        template.set_icon(icon_filepath)
        return template

    def get_all_analysis_templates(self):
        """Get a list of all Analysis Template objects currently available to this connection

        :rtype: list of :class:`~eureqa.analysis_templates.AnalysisTemplate`
        """

        self._session.report_progress('Getting details for analysis_templates.')
        results = self._session.execute('/analysis_templates', 'GET')
        return [AnalysisTemplate(x, self) for x in results]

    @property
    def search_templates(self):
        """Return all Search Templates available to the current connection

        :rtype: SearchTemplates
        """
        return SearchTemplates(self)

    def _list_search_templates(self):
        endpoint = '/fxp/search_templates'
        self._session.report_progress('Getting list of search templates')
        body = self._session.execute(endpoint, 'GET')
        templates = [x['search_template_id'] for x in body]
        return templates

    def _get_analysis_template(self, template_id):
        self._session.report_progress('Getting details for analysis_template: \'%s\'.' % template_id)
        result = self._session.execute('/analysis_templates/%s' % template_id, 'GET')
        return AnalysisTemplate(result, self)

    def _get_analysis_template_by_name(self, template_name):
        self._session.report_progress('Finding first analysis template with name: \'%s\'.' % template_name)
        for t in self.get_all_analysis_templates():
            if (t.name == template_name):
                return t
        raise Exception("No analysis template has name '%s'" % (template_name))

    def _get_search_by_search_id(self, data_source_id, search_id):
        self._session.report_progress('Getting details for search: \'%s\' in datasource: \'%s\'.' % (search_id, data_source_id))
        endpoint = "/fxp/datasources/%s/searches/%s" % (data_source_id, search_id)
        result = self._session.execute(endpoint, 'GET')
        return Search(result, self)

    def _get_solution_by_id(self, data_source_id, search_id, solution_id):
        self._session.report_progress('Getting details for solution: \'%s\' in search: \'%s\' in datasource: \'%s\'.' % (solution_id, search_id, data_source_id))
        endpoint = "/fxp/datasources/%s/searches/%s/solutions/%s" % (data_source_id, search_id, solution_id)
        result = self._session.execute(endpoint, 'GET')
        return Solution(result, self)

    def _get_solutions_by_id(self, data_source_id, search_id):
        self._session.report_progress('Getting details for solutions: \'%s\' in search: \'%s\' in datasource: \'%s\'.' % (solution_id, search_id, data_source_id))
        endpoint = "/fxp/datasources/%s/searches/%s/solutions" % (data_source_id, search_id)
        result = self._session.execute(endpoint, 'GET')
        return [Solution(body, self) for body in result]

    def _get_rest_endpoint_version(self):
        try:
            self._session.report_progress('Getting settings_info.')
            settings = self._session.execute('/api/v2/fxp/settings_info', 'GET')
            return settings['api_version']
        except Http404Exception:
            # The version of eeb that don't expose this endpoint will return this.
            return 0

    def _get_session_key(self):
        return self._session._get_session_key()

    def _verify_version(self):
        server_version = self._get_rest_endpoint_version()
        if Eureqa.API_VERSION != server_version:
            message = 'An incorrect version of the API library is used. ' \
                      'The server version is %s, but the library version is %s.' % (server_version, Eureqa.API_VERSION)
            raise Exception(message)

    def _create_organization(self, name):
        self._session.report_progress('Creating organization: \'%s\'.' % name)
        self._session.execute('/api/v2/organizations', 'POST', args={'name': '%s' % name})
        return _Organization(self, name)

    def _get_organization(self, id):
        self._session.report_progress('Getting organization: \'%s\'.' % id)
        self._session.execute('/api/v2/organizations/%s' % id, 'GET')
        return _Organization(self, id)

    def _delete_organization(self, id):
        self._session.report_progress('Deleting organization: \'%s\'.' % id)
        self._session.execute('/api/v2/organizations/%s' % id, 'DELETE')

    def compute_error_metrics(self, datasource, target_variable, model_expression, template_search=None,
                              variable_options=[], row_weight=None, row_weight_type=None, _data_split="all"):
        """
        Compute the :py:class:`~eureqa.error_metric.ErrorMetrics` for the specified model, against the specified target_variable

        :param DataSource datasource: DataSource to compute error against
        :param str target_variable: Variable (or expression) to compare the model to
        :param str model_expression: Model whose error is to be computed
        :param Search template_search: If specified, inherit variable options from the specified search.  Values specified in :variable_options: take precedence over values in this search; use it for finer-grained control instead of or on top of this argument.
        :param VariableOptionsDict variable_options: Override any default behavior for the specified variables.  If the data contains nulls and no null-handling policy is specified, this method will return an error. A list of :py:class:`~eureqa.variable_options.VariableOptions` may also be provided.
        :param str row_weight: Expression to compute the weight of a row (how much that row contributes to the computed error)
        :param str row_weight_type: The type of expression to use to compute row weight (uniform, target_frequency, variable, or custom_expr)
        :return: The computed error metrics
        :rtype: ErrorMetrics
        """
        # '_data_split' parameter is for internal use.  Non-default valid values are 'training' and 'validation'.

        if isinstance(variable_options, list):
            var_options = [x._to_json() for x in variable_options]
        else:
            var_options = [x._to_json() for x in variable_options.itervalues()]

        args = {
            'row_weight': row_weight,
            'row_weight_type': row_weight_type,
            'lhs_expression': target_variable,
            'rhs_expression': model_expression,
            'data_type': _data_split,
            'variable_options': var_options
        }
        if template_search is not None:
            args["search_id"] = template_search._id
        result = self._session.execute('/fxp/datasources/%s/compute_error_metric' % datasource._data_source_id, 'POST', args=args)
        metrics = ErrorMetrics()
        metrics._from_json(result)
        return metrics

    def evaluate_expression(self, datasource, expressions, template_search=None, variable_options=[], _data_split='all'):
	"""
	Evaluates the provided expression against the specified datasource.  Returns the value of the evaluated computation.

        Example:

            values = eureqa.evaluate_expression(['x','y','x^2']) # where

            values['x']   --> [1,2,3,4]

            values['y']   --> [5,6,7,8]

            values['x^2'] --> [1,4,9,16]

            data = pandas.DataFrame(values)  # convert to pandas.DataFrame

        :param DataSource datasource: DataSource to perform the computation against
        :param str expressions: If only one expression is to be evaluated, that expression.  If multiple expressions are to be evaluated, a list of those expressions.
        :param Search template_search: If specified, inherit variable options from the specified search.  Values specified in :variable_options: take precedence over values in this search; use it for finer-grained control instead of or on top of this argument.
        :param VariableOptionsDict variable_options: Override default variable options directly for particular variables.  Set interpretation of NaN values, outliers, etc.  default behavior is to make no changes to the original data.  By default, missing values are not filled; missing values in input data may result in missing values in corresponding computed values. A list of :py:class:`~eureqa.variable_options.VariableOptions` may also be provided.
        """
        # '_data_split' parameter is for internal use only.  Non-default valid values are 'training' and 'validation'.

        if isinstance(expressions, basestring): expressions = [expressions]

        if isinstance(variable_options, list):
            var_options = [x._to_json() for x in variable_options]
        else:
            var_options = [x._to_json() for x in variable_options.itervalues()]

        # 'datasource' should be a DataSource instance.
        # Allow it to be a string too.  For internal use only.
        data_source_id = (datasource if isinstance(datasource, basestring) else datasource._data_source_id)

        endpoint = '/fxp/datasources/%s/evaluate_expression' % data_source_id

        args = {
            'expression': expressions,
            'data_type': _data_split,
            'variable_options': var_options
        }
        if template_search:
            args['search_id'] = template_search._id

        self._session.report_progress('Evaluating expression on datasource: \'%s\'.' % data_source_id)
        body = self._session.execute(endpoint, 'POST', args)
        values = {expr: [] for expr in expressions}
        for series in body['series_details']:
            used_expression = set()
            for expr, column in zip(expressions, series['values']):
                if expr not in used_expression:
                    values[expr].extend(column)
                    used_expression.add(expr)
        return values

    def _expression_values(self, datasource, x_expr, y_expr):
        """
        Invokes the expression_values endpoint with specified x and y expressions. If the x_expr is a
        datetime column, converts the x values into timestamp strings that can be feed back into the
        backend and parsed as datetime again
        """
        data_source_id = (datasource if isinstance(datasource, basestring) else datasource._data_source_id)
        endpoint = '/fxp/datasources/%s/expression_values?expressions[0][x_values]=%s&expressions[0][y_values]=%s' % (data_source_id, x_expr.strip(), y_expr.strip())
        body = self._session.execute(endpoint, 'GET')
        x_val, y_val = [], []
        for data_point in body[0]['values']:
            x_val.append(data_point['x_value'])
            y_val.append(data_point['y_value'])

        # if x_axis_is_datetime is True, the x values from backend are
        # milliseconds since Unix epoch, convert it into a string that backend
        # can parse back as datetime
        if body[0]['x_axis_is_datetime']:
            x_val = [str(datetime.utcfromtimestamp(x/1000)) for x in x_val]

        return x_val, y_val

    def _grant_user_api_access(self, username, organization, as_role="org_admin"):
        # Function is not public; API may change
        if as_role == "org_admin":
            endpoint = "/api/%(organization)s/auth/user/%(username)s/api_access"
        elif as_role == "user_admin":
            endpoint = "/api/auth/user/%(username)s/api_access/%(organization)s"

        endpoint = endpoint % { 'username': username, 'organization': organization }

        self._session.execute(endpoint, 'PUT', {})

    def _revoke_user_api_access(self, username, organization, as_role="org_admin"):
        # Function is not public; API may change
        if as_role == "org_admin":
            endpoint = "/api/%(organization)s/auth/user/%(username)s/api_access"
        elif as_role == "user_admin":
            endpoint = "/api/auth/user/%(username)s/api_access/%(organization)s"

        endpoint = endpoint % { 'username': username, 'organization': organization }

        self._session.execute(endpoint, 'DELETE', {})

    def _generate_api_key(self, key_name):
        # Function is not public; API may change
        endpoint = "/api/auth/keys"
        result = self._session.execute(endpoint, 'POST', {'name': key_name})
        return result

    def _revoke_api_key_by_id(self, key_id):
        # Function is not public; API may change
        endpoint = "/api/auth/keys/%s" % key_id
        self._session.execute(endpoint, 'DELETE', {})

    def _revoke_api_key_by_name(self, key_name):
        # Function is not public; API may change
        # get all (both) keys and then find the key_id for the key_name
        endpoint = "/api/auth/keys"
        keys = self._session.execute(endpoint, 'GET', {})
        id = next((key['id'] for key in keys if key['name'] == key_name), -1)
        ## send the -1 default if not found so that we get the same Exception as _by_id
        endpoint = "/api/auth/keys/%s" % id
        self._session.execute(endpoint, 'DELETE', {})

    def _list_api_keys(self):
        # Function is not public; API may change
        endpoint = "/api/auth/keys"
        result = self._session.execute(endpoint, 'GET', {})
        return result

    def evaluate_models(self, data, solutions, expressions=[], include_data=False, calculate_error_metrics=False):
        """ Evaluates the provided solutions and model strings on the specified data source.

        :param str data: A path to the file, Python file-like object,
            or an existing Eureqa datasource on which the models have to be evaluated.
        :param list solutions: The list of :py:class:`~eureqa.solution.Solution` objects.
            which have to be evaluated against the datasource.
        :param list expressions: The optional list of expressions (could be variables) to include
            in the output data (optional).
        :param bool include_data: The boolean flag which will indicate whether all data columns will be included into the response.
        :param bool calculate_error_metrics: The boolean flag which will indicate whether the error metrics
            have to be calculated.

        :rtype :py:class:`~eureqa.model_evaluation.ModelEvaluation`
        """

        if isinstance(data, DataSource):
            data = data._duplicate(data.name + " evaluate models", hidden=True)
        else:
            data = self.create_data_source("Evaluate models", data, _hidden=True)
        search = data._apply_solutions(solutions, expressions)
        return search._evaluate_solutions(include_data, calculate_error_metrics)




#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter_value import ParameterValue

import json

class NumericParameterValue(ParameterValue):
    """Numeric parameter value description for analysis template

    :param Eureqa eureqa: A eureqa connection.
    :param str id: The id of the parameter.
    :param float value: The parameter value.

    :var str id: The id of the parameter.
    :var float NumericParameterValue.value: The parameter value.
    """

    def __init__(self, eureqa, id, value):
        """Initializes a new instance of the ~NumericParameterValue class"""
        ParameterValue.__init__(self, id, value, "numeric")

    def _to_json(self):
        body = {}
        ParameterValue._to_json(self, body)
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(eureqa, body, execution=None):
        param = NumericParameterValue(None, None, None)
        ParameterValue._from_json(param, body)
        if param._type != "numeric": raise Exception("Invalid type '%s' specified for numeric parameter value" % param._type)
        return param



from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

class _Table(_Component):
    """ Low level Table component for showing tables within an analyis. See :class:`TableBuilder` for creating tables.
    """

    _component_type_str = 'TABLE'

    def __init__(self, table_id=None, columns=None, defaultSort=None, filters=None, searchBoxAttributes=None,
                 pageSizes=None, searchBoxPlaceholder=None, striped=None, title=None,
                 _analysis=None, _component_id=None, _component_type=None):
        super(_Table, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

        if table_id is not None:
            self._table_id = table_id

        if columns is not None:
            self._columns = columns

        if defaultSort is not None:
            self._defaultSort = defaultSort

        if filters is not None:
            self._filters = filters

        if searchBoxAttributes is not None:
            self._searchBoxAttributes = searchBoxAttributes

        if pageSizes is not None:
            self._pageSizes = pageSizes

        if searchBoxPlaceholder is not None:
            self._searchBoxPlaceholder = searchBoxPlaceholder

        if striped is not None:
            self._striped = striped

        if title is not None:
            self._title = title


    @property
    def table_id(self):
        return self._table_id

    @table_id.setter
    def table_id(self, val):
        self._table_id = val
        self._update()


    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, val):
        self._columns = val
        self._update()


    @property
    def defaultSort(self):
        return self._defaultSort

    @defaultSort.setter
    def defaultSort(self, val):
        self._defaultSort = val
        self._update()


    @property
    def filters(self):
        return self._filters

    @filters.setter
    def filters(self, val):
        self._filters = val
        self._update()


    @property
    def searchBoxAttributes(self):
        return self._searchBoxAttributes

    @searchBoxAttributes.setter
    def searchBoxAttributes(self, val):
        self._searchBoxAttributes = val
        self._update()


    @property
    def pageSizes(self):
        return self._pageSizes

    @pageSizes.setter
    def pageSizes(self, val):
        self._pageSizes = val
        self._update()


    @property
    def searchBoxPlaceholder(self):
        return self._searchBoxPlaceholder

    @searchBoxPlaceholder.setter
    def searchBoxPlaceholder(self, val):
        self._searchBoxPlaceholder = val
        self._update()


    @property
    def striped(self):
        return self._striped

    @striped.setter
    def striped(self, val):
        self._striped = val
        self._update()


    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, val):
        self._title = val
        self._update()



    def _fields(self):
        return super(_Table, self)._fields() + [ 'table_id', 'columns', 'defaultSort', 'filters', 'searchBoxAttributes', 'pageSizes', 'searchBoxPlaceholder', 'striped', 'title' ]



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter

import json

class TopLevelModelParameter(Parameter):
    """TopLevelModel parameter description for analysis template. For selecting a single model (either an existing one within a datasource and search, or a custom one).

    :param str id: The id of the parameter that will be passed together with its value to an analysis script.
    :param str label: The parameter label that will be shown in UI.
    :param bool custom_disabled: Whether to block the user from setting a custom expression

    :var str id: The id of the parameter that will be passed together with its value to an analysis script.
    :var str TopLevelModelParameter.label: The parameter label that will be shown in UI.
    :var bool custom_disabled: Whether to block the user from setting a custom expression
    """

    def __init__(self, id, label, custom_disabled):
        """Initializes a new instance of the ~TopLevelModelParameter class
        """
        Parameter.__init__(self, id, label, "top_level_model")
        self.custom_disabled = custom_disabled

    def _to_json(self):
        body = Parameter._to_json(self)
        if self.custom_disabled is not None:
            body["custom_disabled"] = self.custom_disabled
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(body):
        ret = TopLevelModelParameter(None, None, None)
        Parameter.from_json(ret, body)
        if ret._type != "top_level_model": raise Exception("Invalid type '%s' specified for top_level_model parameter" % ret._type)
        if 'custom_disabled' in body:
            ret.custom_disabled = body["custom_disabled"]
        return ret



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from analysis_template import AnalysisTemplate
from combo_box_parameter import ComboBoxParameter
from combo_box_parameter_value import ComboBoxParameterValue
from data_file_parameter import DataFileParameter
from data_file_parameter_value import DataFileParameterValue
from data_source_parameter import DataSourceParameter
from data_source_parameter_value import DataSourceParameterValue
from execution import Execution
from numeric_parameter import NumericParameter
from numeric_parameter_value import NumericParameterValue
from parameters import Parameters
from parameters_values import ParametersValues
from progress_update import ProgressUpdate
from text_parameter import TextParameter
from text_parameter_value import TextParameterValue
from top_level_model_parameter import TopLevelModelParameter
from top_level_model_parameter_value import TopLevelModelParameterValue
from variable_parameter import VariableParameter
from variable_parameter_value import VariableParameterValue
from parameter_validation_result import ParameterValidationResult

__all__ = ['AnalysisTemplate', 'ComboBoxParameter', 'ComboBoxParameterValue','DataFileParameter', 'DataFileParameterValue',
           'DataSourceParameter', 'DataSourceParameterValue', 'Execution', 'NumericParameter', 'NumericParameterValue',
           'Parameters', 'ParametersValues', 'ProgressUpdate', 'TextParameter', 'TextParameterValue', 'TopLevelModelParameter',
           'TopLevelModelParameterValue', 'VariableParameter', 'VariableParameterValue', 'ParameterValidationResult']



from eureqa.analysis.components.base import _Component

class VariableLink(_Component):
    """
    Create a VariableLink component

    Example of using HtmlBlock with a VariableLink::

        vl = VariableLink(d, 'X')
        h = HtmlBlock('Formatted <b>Card</b> in the Analysis, with reference to {0}'.format(analysis.html_ref(vl)))
        analysis.create_card(h)


    :param DataSource datasource: datasource from which to find the variable
    :param str variable_name: name of the variable
    """

    _component_type_str = 'VARIABLE_LINK'

    def __init__(self, datasource=None, variable_name=None, _analysis=None, _component_id=None, _component_type=None):

        # implementation details:
        # self.datasource is a property, property
        # self._datasource is the actual datasource object (cache of the datasource) which may or may not be fetched
        # self._datasource_id a string, and its setting/handling is handled by the JsonRest implementation'

        if datasource is not None:
            self.datasource = datasource

        if variable_name is not None:
            self._variable_name = variable_name

        super(VariableLink, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def datasource(self):
        """The data source providing data for this component

        :rtype: DataSource
        """
        # Note that the _datasource field is set by the underlying _Component
        # based on the _datasource_id string
        if hasattr(self, "_eureqa"):
            return self._eureqa.get_data_source_by_id(self._datasource_id)

    @datasource.setter
    def datasource(self, val):
        self._datasource_id = val._data_source_id
        self._eureqa = val._eureqa
        self._update()


    @property
    def variable_name(self):
        """The name of the variable from this datasource to link to

        :rtype: str
        """
        return self._variable_name

    @variable_name.setter
    def variable_name(self, val):
        self._variable_name = val
        self._update()

    def _fields(self):
        return super(VariableLink, self)._fields() + [ 'datasource_id', 'variable_name' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from cStringIO import StringIO

class AnalysisFile(object):
    """
    AnalysisFile object

    Represents a file that is attached to an Analysis, and can be
    embedded in various places within the Analysis.
    AnalysisFiles are often either images (.jpg, .png, etc) or
    raw data files (.csv, .xls, etc), but any type of file is supported.

    Do not construct this class directly. Use :meth:`AnalysisFile.create`

    """

    def __init__(self, _analysis, _file_id, _filename=None):
        # Analysis object
        self._analysis = _analysis
        # File ID (string); must be a valid ID of a file on the server
        self._file_id = _file_id
        self._filename = _filename

    @classmethod
    def create(_cls, analysis, file, filename=None):
        """
        Upload a new file to an Eureqa Analysis.

        :param eureqa.analysis.Analysis analysis: Analysis to upload the file to
        :param str file: File to upload.  Can be either a string, in which case it is treated as the binary data of a file, or an actual Python file object or file-like object.
        :param str filename: The default name of file when downloaded
        :rtype: AnalysisFile
        """
        needs_close = False
        if isinstance(file, basestring):
            file = StringIO(file)

        if not filename:
            filename = "file"

        res = analysis._eureqa._session.execute("/analysis/%s/files" % analysis._id,
                                                "POST", files={"file": (filename, file)})

        return _cls(analysis, res["file_id"], filename)

    def update(self, file):
        """
        Update the contents of an existing file on the server.  Overwrites the file's previous data.

        :param str file: File to upload.  Can be either a string, in which case it is treated as the binary data of a file, or an actual Python file object or file-like object.
        """
        needs_close = False
        if isinstance(file, basestring):
            file = StringIO(file)

        self._analysis._eureqa._session.execute("/analysis/%s/files/%s" % (self._analysis._id, self._file_id),
                                                "PUT", raw_data=file)

    def get(self):
        """
        Return the raw data of this file

        :rtype: AnalysisFile.file
        """
        data = self._analysis._eureqa._session.execute("/analysis/%s/files/%s" % (self._analysis._id, self._file_id),
                                                       "GET", raw_return=True)
        return StringIO(data)

    @classmethod
    def _get_for_analysis(cls, analysis):
        """
        :param eureqa.analysis.Analysis analysis: Analysis to consider
        :return: Return a list of all file objects that are present in 'analysis'
        """
        files_json = analysis._eureqa._session.execute("/analysis/%s/files" % (analysis._id), "GET")
        files_list = [cls(analysis, x["file_id"], x.get("file_name")) for x in files_json]
        return files_list

    def delete(self):
        """
        Delete this file from the server
        """
        self._analysis._eureqa._session.execute("/analysis/%s/files/%s" % (self._analysis._id, self._file_id),
                                                "DELETE")

    def url(self):
        """
        :return: URL that this image can be reached from
        """
        return "/api/%(organization)s/analysis/%(analysis_id)s/files/%(file_id)s" % {
            "organization": self._analysis._eureqa._session.organization,
            "analysis_id": self._analysis._id,
            "file_id": self._file_id
        }



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json

class DataSplitting:
    """Represents data splitting settings for the genetic algorithm. Each :py:class:`~eureqa.search.Search` contains a DataSplitting
    object that describes how the genetic algorithm will run. 

    :param bool shuffle: Indicates whether data are shuffled or not before the training.
    :param float training_data_percentage: The percentage of data used for training.
    :param float validation_data_percentage: The percentage of data used for validation.
    :param str training_selection_expression: The expression used for selecting the training set.
    :param str validation_selection_expression: The expression used for selecting the validation set.
    :param str training_selection_expression_type: The type of expression used for selecting the training set. ('EXPRESSION','VARIABLE',None)
    :param str validation_selection_expression_type: The type of expression used for selecting the validation set. ('EXPRESSION','VARIABLE',None)

    :var bool shuffle: Indicates whether data are shuffled or not before the training.
    :var float training_data_percentage: The percentage of data used for training.
    :var float validation_data_percentage: The percentage of data used for validation.
    :var str training_selection_expression: The expression used for selecting the training set.
    :var str validation_selection_expression: The expression used for selecting the validation set.
    :var str training_selection_expression_type: The type of expression used for selecting the training set. ('EXPRESSION','VARIABLE',None)
    :var str validation_selection_expression_type: The type of expression used for selecting the validation set. ('EXPRESSION','VARIABLE',None)
    """

    def __init__(self, shuffle=None, training_data_percentage=None, validation_data_percentage=None,
                 training_selection_expression=None, validation_selection_expression=None,
                 training_selection_expression_type=None, validation_selection_expression_type=None):
        self._type = 'custom'
        self.shuffle = shuffle
        self.training_data_percentage = training_data_percentage
        self.validation_data_percentage = validation_data_percentage
        self.training_selection_expression = training_selection_expression
        self.validation_selection_expression = validation_selection_expression
        self.training_selection_expression_type = training_selection_expression_type
        self.validation_selection_expression_type = validation_selection_expression_type


    def _to_json(self, body):
        # The data splitting parameters are part of the entire search block.
        # So they have to be appended to the existing body.
        # If the user hasn't set any parameters yet, don't write any out yet;
        # we need some information, otherwise this isn't a valid block.
        if self._type == 'custom':
            if self.training_data_percentage and self.validation_data_percentage:
                body['shuffle_rows'] = self.shuffle
                body['training_data_percent'] = self.training_data_percentage
                body['validation_data_percent'] = self.validation_data_percentage
                body['data_splitting'] = self._type
        else:
            body['data_splitting'] = self._type
        if self.training_selection_expression:
            body['training_selection_expression'] = self.training_selection_expression
        if self.validation_selection_expression:
            body['validation_selection_expression'] = self.validation_selection_expression
        if self.training_selection_expression_type:
            body['training_selection_expression_type'] = self.training_selection_expression_type
        if self.validation_selection_expression_type:
            body['validation_selection_expression_type'] = self.validation_selection_expression_type


    def _from_json(self, body):
        self._type = body.get('data_splitting')
        if self._type == 'custom':
            self.shuffle = body.get('shuffle_rows')
            self.training_data_percentage = body.get('training_data_percent')
            self.validation_data_percentage = body.get('validation_data_percent')
        else:
            self.shuffle = None
            self.training_data_percentage = None
            self.validation_data_percentage = None

        self.training_selection_expression = body.get('training_selection_expression')
        self.validation_selection_expression = body.get('validation_selection_expression')
        self.training_selection_expression_type = body.get('training_selection_expression_type')
        self.validation_selection_expression_type = body.get('validation_selection_expression_type')

        self._body = body

    @classmethod
    def from_json(cls, body):
        dataSplitting = DataSplitting(None)
        dataSplitting._from_json(body)
        return dataSplitting

    def __eq__(self, other):
        if not isinstance(other, DataSplitting):
            return False
        return (self._type == other._type and self.shuffle == other.shuffle
                and self.training_data_percentage == other.training_data_percentage
                and self.validation_data_percentage == other.validation_data_percentage)

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        body = {}
        self._to_json(body)
        return json.dumps(body)


_in_order_sample = DataSplitting(shuffle=None, training_data_percentage=None, validation_data_percentage=None)
_in_order_sample._type = "extrapolate"

_random_sample = DataSplitting(shuffle=None, training_data_percentage=None, validation_data_percentage=None)
_random_sample._type = "equal"


def _from_json(body):
    splitting = DataSplitting(None)
    splitting._from_json(body)
    return splitting



import string
from base import _Component

def _gen_component_reference_str(component_id):
    return '<nutonian-component id="{0}" />'.format(component_id)

class _ComponentFormatter(string.Formatter):
    def __init__(self, set_component_id_func, dry_run = False):
        self._dry_run = dry_run
        self._component_definitions = []
        self._set_component_id = set_component_id_func

    def format_field(self, value, format_spec):
        if not isinstance(value, _Component):
            raise TypeError("Argument to format must be a Component")

        # keep track of all the components we have been asked to format
        comp_obj = value
        self._set_component_id(comp_obj)
        self._component_definitions.append(comp_obj._to_json())

        # replace the value to be formatted to a component reference and call parent class implementation
        new_value = _gen_component_reference_str(comp_obj._component_id)
        if self._dry_run:
            del comp_obj._component_id
        return super(_ComponentFormatter, self).format_field(new_value, format_spec)

def _set_component_id_via_placeholder(comp_obj):
    comp_obj._component_id = 'comp_id_placeholder'

class _SetComponentIDViaAnalysis(object):
    def __init__(self, analysis):
        self._analysis = analysis

    def __call__(self, comp_obj):
        comp_obj._associate(self._analysis)

class _SetComponentIDViaTable(object):
    def __init__(self, analysis):
        self._analysis = analysis

    def __call__(self, comp_obj):
        comp_obj._associate_with_table(self._analysis)

class FormattedText(object):
    """ Creates a formatted string, substituing format specifications with component references.

    For example::

        FormattedText("Tooltip {0}", Tooltip(html='Here', tooltip='this is the tooltip'))

    :param str format_str: The string to format


    """

    def __init__(self, format_str, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._format_str = format_str

        # Dry run the format in the c'tor so the format string is checked at construction time
        fmt = _ComponentFormatter(_set_component_id_via_placeholder, dry_run = True)
        self._component_references  = fmt.format(self._format_str, *self._args, **self._kwargs)
        self._component_definitions = fmt._component_definitions

    def _get_component_ref_and_defs(self, analysis, associate_with_analysis):
        if associate_with_analysis:  # only need to re-format if analysis is provided
            set_component_id_via_analysis = _SetComponentIDViaAnalysis(analysis)
            fmt = _ComponentFormatter(set_component_id_via_analysis)
            self._component_references  = fmt.format(self._format_str, *self._args, **self._kwargs)
            self._component_definitions = fmt._component_definitions
        else:
            set_component_id_via_table = _SetComponentIDViaTable(analysis)
            fmt = _ComponentFormatter(set_component_id_via_table)
            self._component_references  = fmt.format(self._format_str, *self._args, **self._kwargs)
            self._component_definitions = fmt._component_definitions

        return self._component_references, self._component_definitions

def _get_component_ref_and_defs_for_value(rendered_value, analysis, associate_with_analysis=True):
    if isinstance(rendered_value, _Component):
        comp = rendered_value
        if associate_with_analysis:
            comp._associate(analysis)
            comp_json = [comp._to_json()]
        else:
            comp_json = comp._associate_with_table(analysis)
        return _gen_component_reference_str(comp._component_id), comp_json
    elif isinstance(rendered_value, FormattedText):
        return rendered_value._get_component_ref_and_defs(analysis, associate_with_analysis)
    else:
        return rendered_value, []



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from data_splitting import DataSplitting
from complexity_weights import ComplexityWeights
from math_block_set import MathBlockSet
import json
import math_block
import math_block_set
from variable_options import VariableOptions
from variable_options_dict import VariableOptionsDict


class SearchSettings(object):
    """A set of settings that describe how a :py:class:`~eureqa.search.Search` will run. To create an instance of this class,
    call a method from :py:meth:`~eureqa.eureqa.Eureqa.search_templates`. After than, the settings can then be
    changed from their default values. Once the desired settings are attained, pass the SearchSettings instance into :py:func:`~eureqa.data_source.DataSource.create_search` to create
    a :py:class:`~eureqa.search.Search` with those settings.

    :var str `~eureqa.search_settings.SearchSettings.name`: The name of the search.
    :var str target_variable: The target variable.
    :var list input_variables: The list of input variables.
    :var MathBlockSet `~eureqa.search_settings.SearchSettings.math_blocks`: The set of mathematical operations allowed in the search's solution.
    :var DataSplitting data_splitting: The data splitting settings for the search algorithm.
    :var str error_metric: The error metric that will be used for this search. Choose from the error metrics in :mod:`~eureqa.error_metric`.
    :var int maximum_history_absolute_rows: The maximum number of rows that can be used in range based functions.
    :var list prior_solutions: The list of prior solutions.
    :var str row_weight: The row weight expression.
    :var str row_weight_type: The row weight type expression.
    :var str target_expression: The target expression.
    :var VariableOptionsDict variable_options: Override default behavior for the specified variables.
    :var int default_min_delay: the default minimum number of rows each variable will be delayed in the model, if not overridden by a specific variable option.

    Example::

      # create a new SearchSettings object with the default settings for a numeric search
      settings = e.search_templates.numeric("Sample Model", target_variable, variables - {target_variable})

      # enable the pow building block and set its complexity to 3
      settings.math_blocks.pow.enable(complexity=3)
      # disable the pow building block in searches
      settings.math_blocks.pow.disable()
      # set the complexity of log to 5 (do not change enabled or disabled)
      settings.math_blocks.log.complexity = 5

      # enable a list of blocks with default complexity
      for block in [settings.math_blocks.pow, settings.math_blocks.exp]:
          block.enable()

      # print a list of enabled blocks w/ complexity
      for block in settings.math_blocks:
          if block.enabled:
               print block.name + \": \" + str(block.complexity)

      # select a different error metric
      settings.error_metric = error_metric.mean_square_error()

      # create a new Search with these settings
      search = my_data_source.create_search(settings)


   """

    def __init__(self, eureqa=None,
                 name=None,
                 search_template_id=None,
                 target_variable=None,
                 input_variables=None,
                 data_splitting=None,
                 error_metric=None,
                 max_num_variables_per_term=None,
                 prior_solutions=None,
                 target_expression=None,
                 row_weight_type=None,
                 row_weight=None,
                 maximum_history_percentage=None,
                 maximum_history_absolute_rows=None,
                 default_min_delay=None,
                 variable_options=None,
                 datasource_id=None):
        """For internal use only.
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._body = {}
        self._eureqa = eureqa
        self.name = name
        self._search_template_id = search_template_id
        self._target_variable = target_variable
        self._input_variables = UpdatingList(self._update, input_variables or [])
        self.math_blocks = MathBlockSet()
        self._complexity_weights = ComplexityWeights()
        self.data_splitting = data_splitting or DataSplitting()
        self.error_metric = error_metric
        self._max_num_variables_per_term = max_num_variables_per_term
        self._maximum_history_percentage = maximum_history_percentage
        self.maximum_history_absolute_rows = maximum_history_absolute_rows
        self.prior_solutions = prior_solutions
        self._row_weight = row_weight or 1.0
        self._row_weight_type = row_weight_type or 'uniform'
        if target_expression is not None:
            self.target_expression = target_expression
        else:
            # No target expression defined.
            # Don't trigger the logic in the target_expression setter to use a custom expression.
            self._target_expression = None
        self._default_min_delay = default_min_delay
        if variable_options:
            self._variable_options = VariableOptionsDict([(x.name, x) for x in variable_options])
        else:
            self._variable_options = VariableOptionsDict([(x, VariableOptions(name=x, min_delay=default_min_delay, _update_parent_settings=self._update)) for x in (input_variables or [])])
        self._datasource_id = datasource_id
        if self._row_weight_type == 'custom':
            self._row_weight_type = 'custom_expr'


    @property
    def target_variable(self):
        return self._target_variable
    @target_variable.setter
    def target_variable(self, val):
        self._target_variable = val
        self._update()


    @property
    def input_variables(self):
        return self._input_variables
    @input_variables.setter
    def input_variables(self, val):
        self._input_variables = val
        self._update()

    @property
    def variable_options(self):
        return self._variable_options

    @property
    def default_min_delay(self):
        """
        The default minimum number of rows each variable will be delayed
        in the model, if not overridden by a specific variable
        option. This value is the number of rows the resulting
        model can predict into the future.
        """
        return self._default_min_delay
    @default_min_delay.setter
    def default_min_delay(self, val):
        self._default_min_delay = val
        self._update()

    @property
    def math_blocks(self):
        """
        The :class:`~eureqa.math_block_set.MathBlockSet` which represents mathematical operations allowed in solutions to the search.
        """
        return self._math_blocks
    @math_blocks.setter
    def math_blocks(self, val):
        self._math_blocks = val

    @property
    def target_expression(self):
        """ The target expression to optimize.

        Only set if a custom expression is required.
        If set, the resulting search will be treated as an "Advanced"
        search in the UI; the variable chooser will be replaced with
        an expression editor.
        """
        return self._target_expression
    @target_expression.setter
    def target_expression(self, val):
        self._target_expression = val
        self._body['target_expression_edited'] = True
    @target_expression.deleter
    def target_expression(self):
        # Deleting a custom target_expression reverts to the default
        assert hasattr(self, '_original_target_expression'), "Can't delete the target expression:  Was not created from a template; no default to fall back to."
        self._target_expression = self._original_target_expression
        if 'target_expression_edited' in self._body:
            del self._body['target_expression_edited']

    @property
    def row_weight(self):
        """The relative weight that each row is given when computing the error
        metric values for a model.

        See description of :attr:`~row_weight_type` for
        restrictions on the value of row_weight.

        """
        return self._row_weight
    @row_weight.setter
    def row_weight(self, val):
        self._row_weight = val

    @property
    def row_weight_type(self):
        """The method to use for weighting the error metric calculation.

        Options are 'uniform' (default), 'target_frequency',
        'variable', and 'custom_expr'.  If set to 'variable',
        :attr:`~row_weight` must be the name of a variable. If set to
        'custom_expr', :attr:`~row_weight` may be an arbitrary
        expression.

        """
        return self._row_weight_type
    @row_weight_type.setter
    def row_weight_type(self, val):
        self._row_weight_type = val


    @classmethod
    def from_json(cls, eureqa, body):
        name = body.get('search_name')
        datasource_id = body.get('datasource_id')
        search_template_id = body.get('search_template_id')
        target_variable = body.get('target_variable')
        input_variables = body.get('input_variables')
        target_expression = body.get('target_expression')
        error_metric = body.get('error_metric')
        max_num_variables_per_term = body.get('max_num_variables_per_term')
        row_weight = body.get('row_weight')
        row_weight_type = body.get('row_weight_type')
        data_splitting = DataSplitting.from_json(body)
        maximum_history_absolute_rows = body.get('max_delay', None)
        maximum_history_percentage = body.get('maximum_history_percent', None)
        prior_solutions = body.get('prior_solutions', [])
        default_min_delay = body.get('default_min_delay', None)
        ss = SearchSettings(eureqa=eureqa,
                            name=name,
                            search_template_id=search_template_id,
                            target_variable=target_variable,
                            input_variables=input_variables,
                            data_splitting=data_splitting,
                            error_metric=error_metric,
                            max_num_variables_per_term=max_num_variables_per_term,
                            prior_solutions=prior_solutions,
                            row_weight=row_weight,
                            row_weight_type=row_weight_type,
                            target_expression=target_expression,
                            maximum_history_percentage=maximum_history_percentage,
                            maximum_history_absolute_rows=maximum_history_absolute_rows,
                            default_min_delay=default_min_delay,
                            variable_options = [VariableOptions._from_json(var) for var in (body.get('variable_options') or [])],
                            datasource_id=datasource_id)
        ss.math_blocks = math_block_set.MathBlockSet.from_json(body.get('building_blocks') or [])
        ss._original_target_expression = target_expression
        ss._body['target_expression_edited'] = body.get('target_expression_edited')
        for var_opt in ss.variable_options.keys():
            ss.variable_options[var_opt]._update_parent_settings = ss._update
        return ss

    def _to_json(self):
        if self.name is not None:
            self._body['search_name'] = self.name

        if self._search_template_id is not None:
            self._body['search_template_id'] = self._search_template_id

        if self.target_variable is not None:
            self._body['target_variable'] = self.target_variable

        if self.input_variables is not None:
            self._body['input_variables'] = self.input_variables

        if self.target_expression is not None:
            self._body['target_expression'] = self.target_expression

        if self.error_metric is not None:
            self._body['error_metric'] = self.error_metric

        if self._max_num_variables_per_term is not None:
            self._body['max_num_variables_per_term'] = self._max_num_variables_per_term

        if self.row_weight is not None:
            self._body['row_weight'] = self.row_weight

        if self.row_weight_type is not None:
            self._body['row_weight_type'] = self.row_weight_type

        self.data_splitting._to_json(self._body)

        if self.maximum_history_absolute_rows is not None:
            self._body['max_delay'] = self.maximum_history_absolute_rows
        elif self._maximum_history_percentage is not None:
            self._body['maximum_history_percent'] = self._maximum_history_percentage

        self._body['building_blocks'] = self.math_blocks._to_json()
        self._body['complexity_weights'] = self._complexity_weights._to_json()
        self._body['variable_options'] = self.variable_options._to_json()

        if self._datasource_id is not None:
            self._body['datasource_id'] = self._datasource_id

        if self.prior_solutions is not None:
            self._body['prior_solutions'] = self.prior_solutions

        if self.default_min_delay is not None:
            self._body['default_min_delay'] = self.default_min_delay

        return self._body

    def _from_json(self, body):

        self._body = {}
        self.name = body.get('search_name')
        self._datasource_id = body.get('datasource_id')
        self._search_template_id = body.get('search_template_id')
        self._target_variable = body.get('target_variable')
        self._input_variables = UpdatingList(self._update, body.get('input_variables'))
        self.math_blocks = math_block_set.MathBlockSet.from_json(body.get('building_blocks'))
        self._complexity_weights = ComplexityWeights()
        self.data_splitting = DataSplitting.from_json(body)
        self.error_metric = body.get('error_metric')
        self._max_num_variables_per_term = body.get('max_num_variables_per_term')
        self._maximum_history_percentage = body.get('maximum_history_percent', None)
        self.maximum_history_absolute_rows = body.get('max_delay', None)
        self.prior_solutions = body.get('prior_solutions', [])
        self._row_weight = body.get('row_weight')
        self._row_weight_type = body.get('row_weight_type')
        self._target_expression = body.get('target_expression')
        self._default_min_delay = body.get('default_min_delay', None)

        if body.get('target_expression_edited'):
            self._body['target_expression_edited'] = True
            if body.get('target_expression_from_template') is not None:
                self._original_target_expression = body.get('target_expression_from_template')
        else:
            if 'target_expression_edited' in self._body:
                del self._body['target_expression_edited']
            self._original_target_expression = body.get('target_expression')

        var_opts = [VariableOptions._from_json(var_opt, self._update) for var_opt in body.get('variable_options')]
        self._variable_options = VariableOptionsDict((var.name, var) for var in var_opts)

    def __str__(self):
        return json.dumps(self._to_json())

    # used to regenerate target_expression when a property that affects the target expression is changed
    # also checks the SearchSettings object is valid
    def _update(self):
        endpoint = '/fxp/search_templates/%s/create_settings' % self._search_template_id
        resp = self._eureqa._session.execute(endpoint, 'POST', self._to_json())
        self._from_json(resp)

#Helper class to allow SearchSettings to update on input_variables change
class UpdatingList(list):

    def __init__(self, update_parent_settings, *args, **kwargs):
        super(UpdatingList, self).__init__(*args, **kwargs)
        self.update_parent_settings = update_parent_settings

    def __setitem__(self, key, value):
        super(UpdatingList, self).__setitem__(key, value)
        if hasattr(self, 'update_parent_settings'):
            self.update_parent_settings()

    def __delitem__(self, key):
        super(UpdatingList, self).__delitem__(key)
        if hasattr(self, 'update_parent_settings'):
            self.update_parent_settings()

    def append(self, val):
        super(UpdatingList, self).append(val)
        if hasattr(self, 'update_parent_settings'):
            self.update_parent_settings()



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class Parameter(object):
    """Base class for all analysis templates parameters

    :param str id: The unique identifier for this Parameter.
    :param str label: The human-readable label to be associated with this Parameter.
    """

    def __init__(self, id, label, _type):
        self.id = id
        self.label = label
        self._type = _type
        self._body = {}
        
    def _to_json(self):
        body = self._body
        body['id'] = self.id
        body['label'] = self.label
        body['type'] = self._type
        return body

    def from_json(self, body):
        self.id = body['id']
        self.label = body['label']
        self._type = body['type']
        self._body = body
        
    @staticmethod
    def _from_json(body):
        p = Parameter(None, None, None)
        p.from_json(body)
        return p




# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime

import json

class ParameterValidationResult:
    """Represents an analysis template execution validation result to the server

    :param str type: If this result is INFO, WARNING, or ERROR
    :param str message: The the message
    :param str details: (optional) Detailed message about the result
    :param str parameter_id: (optional) the analysis template parameter that this progress update refers to

    :var str type: If this result is INFO, WARNING, or ERROR
    :var str message: The the message
    :var str details: (optional) Detailed message about the result
    :var str parameter_id: (optional) the analysis template parameter that this progress update refers to
    """

    def __init__(self, type, message=None, details=None, parameter_id=None):
        """For internal use only"""
        self.type     = type
        self.message  = message
        self.details = details
        self.parameter_id = parameter_id

    def _from_json(self, body):
        """For internal use only -- sets from json message body"""
        self.type         = body.get('type')
        self.message      = body.get('message')
        self.details      = body.get('details')
        self.parameter_id = body.get('parameter_id')


    def _to_json(self):
        body = {}

        if not hasattr(self, 'type'):
            raise ValueError('Must specify a type of parameter validation result')
        if not self.type in {'INFO', 'WARNING', 'ERROR'}:
            raise ValueError('type must be INFO, WARNING, ERROR, not ' + str(self.type))
        body['type'] = self.type;

        if (not self.message):
            raise ValueError('message must be specified in parameter validation result')

        body['message'] = self.message

        if (self.details):
            body['details'] = self.details
        if (self.parameter_id):
            body['parameter_id'] = self.parameter_id

        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)



#!/usr/bin/env python

# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This is a companion script for "run_analysis_template.py".
# Their implementations overlap substantially.

import sys
import json
import argparse
import os
import re

if __package__:
    # We're being run as "python -m eureqa install"
    from . import Eureqa
    from .analysis_templates import Parameters
    from .utils._analysis_template_cli import *
    from .session import Http404Exception
else:
    # We're being run as "eureqa/install_analysis_template.py"
    sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
    from eureqa import Eureqa
    from eureqa.analysis_templates import Parameters
    from eureqa.utils._analysis_template_cli import *
    from eureqa.session import Http404Exception

def main(argv):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION, epilog=textwrap.dedent("""\
        Usage Examples:

        To upload the example module included with this project, do:
        python -m eureqa install_analysis_template example_module
        You will be prompted for username and password if necessary.
        Re-run to update the new module in place.

        To install with a custom name and within a Eureqa Organization:
        python -m eureqa install_analysis_template.py --name 'My Template' -u <username> -p <password> -o <organization> example_module

        To overwrite an existing module with ID 52 with the example:
        python -m eureqa install_analysis_template.py --id 52 example_module

        To use as part of a non-interactive shell script (Mac/Linux):
        NEW_TEMPLATE_ID=$(python -m eureqa install_analysis_template -q -u <username> example_module)
        This will use your saved password from the Eureqa API credential store,
        or error out if it isn't available.  (Omit '-q' to instead prompt
        for a password.)
        """))

    ## Pick the command-line arguments that this app needs
    add_module_args(parser)
    add_server_args(parser)
    add_template_args(parser)
    add_script_args(parser)

    ## parse_args() wants the arguments to the command.
    ## argv[0] *is* the command.
    args = parser.parse_args(argv[1:])

    for module_dir in [args.module_dir] + args.dependency_dir:
        module_name = os.path.basename(module_dir)
        if not re.match("[A-Za-z_][A-Za-z0-9_]*", module_name):
            raise RuntimeError, "Invalid directory name: '%s'  Module directory names must contain only [A-Za-z0-9_].  Its first character cannot be a number." % module_name

    params = merge_defaults_from_json(args.module_dir, args)

    clean_and_populate_params(params, not args.quiet)

    if not args.quiet:
        params_no_password = dict(params)
        params_no_password['password'] = '********'
        print "Installing module '%s':\n%s" % (module_name, json.dumps(params_no_password,
                                                                       indent=4,
                                                                       sort_keys=True))

    ## Make sure the module is in path so it can be imported
    mod_path = os.path.dirname(os.path.abspath(params['module_dir']))
    sys.path.insert(0, mod_path)

    parameters = get_parameters_from_module(module_name_from_module_dir(params['module_dir']))._to_json()

    id_ = apply_module(module_name = os.path.basename(args.module_dir),
                       id_=params.get('id', None),
                       parameters=parameters,
                       **params)

    if not args.quiet:
        print "Successfully uploaded to analysis template '%s' (%s)" % (params['name'], id_)
    else:
        print id_

    params['id'] = id_

    if args.save:
        write_defaults_to_json(params, params['module_dir'])

    sys.path.remove(mod_path)


def apply_module(module_dir, module_name, organization, id_, name, description, parameters, url, username, verify_ssl_certificate, dependency_dir, quiet=False, password=False, save_credentials=False, icon_path=None, **unused_args):
    """Create and/or fetch the analysis template as necessary,
    and upload the specified module to it.

    :param str module_dir: The directory where the module to apply can be found.
    :param str module_name: The name of the module to apply.
    :param str organization: The organization containing the data.
    :param str id_: The id for the module.
    :param str name: A name for the template.
    :param str description: A user-defined description for the template.
    :param dict parameters: Template parameters as dictionary.
    :param str url: The url defining the endpoint.
    :param str username: The username for authentication.
    :param bool verify_ssl_certificate: Whether to use certificate-based authentication.
    :param str dependency_dir: A directory where required modules will be found.
    :param bool quiet: Whether to use interactive mode.
    :param bool password: Whether password authentication is to be used.
    :param bool save_credentials: Whether the credentials should be saved by Eureqa.
    :param bool save_credentials: Whether the credentials should be saved by Eureqa. 
    :param str icon_path: path to an icon to use for the analysis
    """

    eureqa_args = {
        'url': url,
        'user_name': username,
        'organization': organization
        }
    if password:
        eureqa_args['password'] = password
    if save_credentials:
        eureqa_args['save_credentials'] = save_credentials

    eureqa_args['verify_ssl_certificate'] = verify_ssl_certificate == 'true'
    eureqa_args['interactive_mode'] = not quiet

    conn = Eureqa(**eureqa_args)

    # if we were given a pre-existing id to use, try to do so
    template = None
    if id_:
        try:
            template = conn._get_analysis_template(id_)
            template.name = name
            template.description = description
            template.parameters = parameters
            template.set_icon(icon_path)
        except Http404Exception:
            if not quiet:
                print "ID '%s' was provided but no analysis template with such ID exists." % id_
                print 'Therefore ID will be ignored a new analysis template will be created.'

    if not template:
        template = conn.create_analysis_template(name, description,
                                                 Parameters._from_json(parameters) if parameters else Parameters(),
                                                 icon_filepath=icon_path)
        id_ = template._id

    template.set_module(module_name, module_dir, ignore_files=[EUREQA_CONFIG_FILE], additional_modules_paths=dependency_dir, icon_path=icon_path)
    return id_


## If we're being run, do stuff.
## (If we're being imported as a library, don't do stuff.)
if __name__ == "__main__":
    main(sys.argv)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

import warnings

class DistributionPlot(_Component):
    """Represents a distribution plot card on the server.

    :param eureqa.DataSource datasource: The data source to which the variable belongs.
    :param str variable_name: The name of the variable that will be displayed on the card.

    :var str ~eureqa.analysis_cards.DistributionPlotCard.title: The card title.
    :var str ~eureqa.analysis_cards.DistributionPlotCard.datasource: The datasource used by the card.
    :var str ~eureqa.analysis_cards.DistributionPlotCard.variable: The variable plotted by the coard.
    """

    _component_type_str = 'VARIABLE_DISTRIBUTION_PLOT'

    def __init__(self, datasource=None, variable_name=None, _analysis=None, _component_id=None, _component_type=None):
        if datasource is not None:
            self.datasource = datasource

        if variable_name is not None:
            self._variable_name = variable_name

        super(DistributionPlot, self).__init__(_analysis=_analysis, _component_id=_component_id,
                                                       _component_type=_component_type)

    @property
    def datasource(self):
        """The data source providing data for this component

        :return: data source providing data for this component
        """
        if hasattr(self, "_eureqa"):
            return self._eureqa.get_data_source_by_id(self._datasource_id)

    @datasource.setter
    def datasource(self, val):
        self._datasource_id = val._data_source_id
        self._eureqa = val._eureqa
        self._update()

    @property
    def variable_name(self):
        """
        Name of the variable to plot

        :rtype: str
        """
        return self._variable_name

    @variable_name.setter
    def variable_name(self, val):
        self._variable_name = val
        self._update()


    # For backwards compatibility
    @property
    def variable(self):
        """
        (Deprecated) Name of the variable to plot.  For backwards compatibility.

        :rtype: str
        """
        warnings.warn("Field name 'variable' on DistributionPlot is deprecated; please use 'variable_name'", DeprecationWarning)
        return self.variable_name

    @variable.setter
    def variable(self, val):
        warnings.warn("Field name 'variable' on DistributionPlot is deprecated; please use 'variable_name'", DeprecationWarning)
        self.variable_name = val


    def _fields(self):
        return super(DistributionPlot, self)._fields() + [ 'datasource_id', 'variable_name' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .math_block import MathBlock
import copy

class MathBlockSet(object):
    """
    Contains a set of :py:class:`~eureqa.math_block.MathBlock` objects that represents a 
    set of all available mathematical operations. Created automatically during the 
    construction of a :py:class:`~eureqa.search_settings.SearchSettings`.

    Every :py:class:`~eureqa.search.Search` and :py:class:`~eureqa.search_settings.SearchSettings` contains
    a MathBlockSet in :py:attr:`eureqa.search.Search.math_blocks` or :py:attr:`eureqa.search_settings.SearchSettings.math_blocks`.
    Use the properties of this variable to access individual
    :py:class:`~eureqa.math_block.MathBlock` objects. The enabled :py:class:`~eureqa.math_block.MathBlock`
    objects of :py:attr:`eureqa.search.Search.math_blocks` determine the operations allowed
    in solutions to the :py:class:`~eureqa.search.Search` .
    """

    def __init__(self):
        self._blocks = copy.deepcopy(self._all_blocks)

    def __iter__(self):
        return (block for block in self._blocks.itervalues() if block.enabled)

    def _to_json(self):
        # Special case so that, if the user hasn't selected any blocks,
        # tell the backend to use default values.
        if len(list(self)) == 0:
            return []

        return [block._to_json() for block in self._blocks.itervalues()]

    def __str__(self):
        from json import dumps
        return dumps(self._to_json(), indent=4, sort_keys=True)

    def __repr__(self):
        return "MathBlockSet(%s)" % repr(list(self))

    @classmethod
    def from_json(cls, blocks):
        math_blocks = MathBlockSet()
        for block in math_blocks:
            block.disable()
        math_blocks._from_json(blocks)
        return math_blocks

    def _from_json(self, blocks_list):
        for jsonblock in blocks_list:
            block = MathBlock.from_json(jsonblock)
            default_block = self._all_blocks.get(block.name)
            block._default_complexity = default_block._complexity if default_block else 4

            ## Resolve name collisions by picking the lower (original) ID
            if block.name not in self._blocks or block._id <= self._blocks[block.name]._id:
                self._blocks[block.name] = block

    def __eq__(self, other):
        ## Two MathBlockSets are equal if they have the same enabled MathBlocks.
        ## "set()" takes an iterable as an argument and calls its __iter__ method
        ## to get the elements to put into the set.
        ## Our __iter__ method filters out disabled MathBlocks; see above.
        ## Note that MathBlock implements all the methods required in order to
        ## be an element in a set(), so this is safe.
        return set(self) == set(other)

    def _get_block(self, name):
        block = self._blocks[name]
        assert name == block.name, "Internal inconsistency:  Index entry for '%s' found block of type '%s'" % (name, block.name)
        return block

    @property
    def const(self):
        """Returns math block for Constant

        :rtype: MathBlock
        """
        return self._get_block('Constant')

    @property
    def int_const(self):
        """Integer Constant math block

        :rtype: MathBlock
        """
        return self._get_block('Integer Constant')


    @property
    def var(self):
        """Input Variable math block

        :rtype: MathBlock
        """
        return self._get_block('Input Variable')

    @property
    def add(self):
        """Addition math block

        :rtype: MathBlock
        """
        return self._get_block('Addition')

    @property
    def sub(self):
        """Subtraction math block

        :rtype: MathBlock
        """
        return self._get_block('Subtraction')

    @property
    def mult(self):
        """Multiplication math block

        :rtype: MathBlock
        """
        return self._get_block('Multiplication')

    @property
    def div(self):
        """Division math block

        :rtype: MathBlock
        """
        return self._get_block('Division')


    @property
    def sin(self):
        """Sine math block

        :rtype: MathBlock
        """
        return self._get_block('Sine')


    @property
    def cos(self):
        """Cosine math block

        :rtype: MathBlock
        """
        return self._get_block('Cosine')

    @property
    def neg(self):
        """Negation math block

        :rtype: MathBlock
        """
        return self._get_block('Negation')

    @property
    def fact(self):

        """Factorial math block

        :rtype: MathBlock
        """
        return self._get_block('Factorial')

    @property
    def pow(self):
        """Power math block

        :rtype: MathBlock
        """
        return self._get_block('Power')

    @property
    def exp(self):
        """Exponential math block

        :rtype: MathBlock
        """
        return self._get_block('Exponential')

    @property
    def log(self):
        """Natural Logarithm math block

        :rtype: MathBlock
        """
        return self._get_block('Natural Logarithm')

    @property
    def abs(self):
        """Absolute Value math block

        :rtype: MathBlock
        """
        return self._get_block('Absolute Value')

    @property
    def if_op(self):
        """If-Then-Else math block

        :rtype: MathBlock
        """
        return self._get_block('If-Then-Else')

    @property
    def logistic(self):
        """Logistic Function math block

        :rtype: MathBlock
        """
        return self._get_block('Logistic Function')

    @property
    def step(self):
        """Step Function math block

        :rtype: MathBlock
        """
        return self._get_block('Step Function')

    @property
    def sign(self):
        """Sign Function math block

        :rtype: MathBlock
        """
        return self._get_block('Sign Function')

    @property
    def gauss(self):
        """Gaussian Function math block

        :rtype: MathBlock
        """
        return self._get_block('Gaussian Function')

    @property
    def min(self):
        """Minimum math block

        :rtype: MathBlock
        """
        return self._get_block('Minimum')

    @property
    def max(self):
        """Maximum math block

        :rtype: MathBlock
        """
        return self._get_block('Maximum')

    @property
    def mod(self):
        """Modulo math block

        :rtype: MathBlock
        """
        return self._get_block('Modulo')

    @property
    def floor(self):
        """Floor math block

        :rtype: MathBlock
        """
        return self._get_block('Floor')

    @property
    def ceiling(self):
        """Ceiling math block

        :rtype: MathBlock
        """
        return self._get_block('Ceiling')

    @property
    def round(self):
        """Round math block

        :rtype: MathBlock
        """
        return self._get_block('Round')

    @property
    def tan(self):
        """Tangent math block

        :rtype: MathBlock
        """
        return self._get_block('Tangent')

    @property
    def equal(self):
        """Equal-To math block

        :rtype: MathBlock
        """
        return self._get_block('Equal-To')

    @property
    def less(self):
        """Less-Than math block

        :rtype: MathBlock
        """
        return self._get_block('Less-Than')

    @property
    def less_equal(self):
        """Less-Than-Or-Equal math block

        :rtype: MathBlock
        """
        return self._get_block('Less-Than-Or-Equal')

    @property
    def greater(self):
        """Greater-Than math block

        :rtype: MathBlock
        """
        return self._get_block('Greater-Than')

    @property
    def greater_equal(self):
        """Greater-Than-Or-Equal math block

        :rtype: MathBlock
        """
        return self._get_block('Greater-Than-Or-Equal')

    @property
    def and_op(self):
        """Logical And math block

        :rtype: MathBlock
        """
        return self._get_block('Logical And')

    @property
    def or_op(self):
        """Logical Or math block

        :rtype: MathBlock
        """
        return self._get_block('Logical Or')

    @property
    def xor(self):
        """Logical Xor math block

        :rtype: MathBlock
        """
        return self._get_block('Logical Xor')

    @property
    def not_op(self):
        """Logical Not math block

        :rtype: MathBlock
        """
        return self._get_block('Logical Not')

    @property
    def tanh(self):
        """Hyperbolic Tangent math block

        :rtype: MathBlock
        """
        return self._get_block('Hyperbolic Tangent')

    @property
    def sqrt(self):
        """Square Root math block

        :rtype: MathBlock
        """
        return self._get_block('Square Root')

    @property
    def delay(self):
        """Delayed Variable math block

        :rtype: MathBlock
        """
        return self._get_block('Delayed Variable')

    @property
    def simple_moving_average(self):
        """Simple Moving Average math block

        :rtype: MathBlock
        """
        return self._get_block('Simple Moving Average')

    @property
    def asin(self):
        """Arcsine math block

        :rtype: MathBlock
        """
        return self._get_block('Arcsine')

    @property
    def acos(self):
        """Arccosine math block

        :rtype: MathBlock
        """
        return self._get_block('Arccosine')

    @property
    def atan(self):
        """Arctangent math block

        :rtype: MathBlock
        """
        return self._get_block('Arctangent')

    @property
    def two_args_atan(self):
        """Two-Argument Arctangent math block

        :rtype: MathBlock
        """
        return self._get_block('Two-Argument Arctangent')

    @property
    def error(self):
        """Error Function math block

        :rtype: MathBlock
        """
        return self._get_block('Error Function')

    @property
    def complementary_error(self):
        """Complementary Error Function math block

        :rtype: MathBlock
        """
        return self._get_block('Complementary Error Function')

    @property
    def weighted_moving_average(self):
        """Weighted Moving Average math block

        :rtype: MathBlock
        """
        return self._get_block('Weighted Moving Average')

    @property
    def _modified_moving_average(self):
        """Modified Moving Average math block

        :rtype: MathBlock
        """
        return self._get_block('Modified Moving Average')

    @property
    def simple_moving_median(self):
        """Simple Moving Median math block

        :rtype: MathBlock
        """
        return self._get_block('Simple Moving Median')

    @property
    def sinh(self):
        """Hyperbolic Sine math block

        :rtype: MathBlock
        """
        return self._get_block('Hyperbolic Sine')

    @property
    def cosh(self):
        """Hyperbolic Cosine math block

        :rtype: MathBlock
        """
        return self._get_block('Hyperbolic Cosine')

    @property
    def asinh(self):
        """Inverse Hyperbolic Sine math block

        :rtype: MathBlock
        """
        return self._get_block('Inverse Hyperbolic Sine')

    @property
    def acosh(self):
        """Inverse Hyperbolic Cosine math block

        :rtype: MathBlock
        """
        return self._get_block('Inverse Hyperbolic Cosine')

    @property
    def atanh(self):
        """Inverse Hyperbolic Tangent math block

        :rtype: MathBlock
        """
        return self._get_block('Inverse Hyperbolic Tangent')


## Do some pre-computation; avoid generating this dictionary repeatedly
def _generate_math_blocks_dict():
    blocks = [
        MathBlock(0,  u'Constant',                     0, u'c_'),
        MathBlock(1,  u'Integer Constant',             1, u'n_'),
        MathBlock(2,  u'Input Variable',               1, u'x_'),
        MathBlock(3,  u'Addition',                     0, u' + '),
        MathBlock(4,  u'Subtraction',                  0, u' - '),
        MathBlock(5,  u'Multiplication',               0, u'*'),
        MathBlock(6,  u'Division',                     2, u'/'),
        MathBlock(7,  u'Sine',                         4, u'sin'),
        MathBlock(8,  u'Cosine',                       4, u'cos'),
        MathBlock(9,  u'Negation',                     4, u'-'),
        MathBlock(10, u'Factorial',                    4, u'factorial'),
        MathBlock(11, u'Power',                        4, u'^'),
        MathBlock(12, u'Exponential',                  4, u'exp'),
        MathBlock(13, u'Natural Logarithm',            2, u'log'),
        MathBlock(14, u'Absolute Value',               4, u'abs'),
        MathBlock(15, u'If-Then-Else',                 1, u'if'),
        MathBlock(23, u'Logistic Function',            4, u'logistic'),
        MathBlock(24, u'Step Function',                2, u'step'),
        MathBlock(25, u'Sign Function',                4, u'sgn'),
        MathBlock(26, u'Gaussian Function',            4, u'gauss'),
        MathBlock(27, u'Minimum',                      1, u'min'),
        MathBlock(28, u'Maximum',                      1, u'max'),
        MathBlock(29, u'Modulo',                       4, u'mod'),
        MathBlock(30, u'Floor',                        4, u'floor'),
        MathBlock(31, u'Ceiling',                      4, u'ceil'),
        MathBlock(32, u'Round',                        4, u'round'),
        MathBlock(33, u'Tangent',                      4, u'tan'),
        MathBlock(34, u'Equal-To',                     4, u'equal'),
        MathBlock(35, u'Less-Than',                    1, u'less'),
        MathBlock(36, u'Less-Than-Or-Equal',           4, u'less_or_equal'),
        MathBlock(37, u'Greater-Than',                 4, u'greater'),
        MathBlock(38, u'Greater-Than-Or-Equal',        4, u'greater_or_equal'),
        MathBlock(39, u'Logical And',                  4, u'and'),
        MathBlock(40, u'Logical Or',                   4, u'or'),
        MathBlock(41, u'Logical Xor',                  4, u'xor'),
        MathBlock(42, u'Logical Not',                  4, u'not'),
        MathBlock(43, u'Hyperbolic Tangent',           4, u'tanh'),
        MathBlock(44, u'Square Root',                  1, u'sqrt'),
        MathBlock(45, u'Delayed Variable',             4, u'delay'),
        MathBlock(46, u'Simple Moving Average',        4, u'sma'),
        MathBlock(49, u'Arcsine',                      4, u'asin'),
        MathBlock(50, u'Arccosine',                    4, u'acos'),
        MathBlock(51, u'Arctangent',                   4, u'atan'),
        MathBlock(52, u'Two-Argument Arctangent',      4, u'atan2'),
        MathBlock(70, u'Error Function',               4, u'erf'),
        MathBlock(71, u'Complementary Error Function', 4, u'erfc'),
        MathBlock(72, u'Weighted Moving Average',      4, u'wma'),
        MathBlock(73, u'Modified Moving Average',      4, u'mma'),
        MathBlock(74, u'Simple Moving Median',         4, u'smm'),
        MathBlock(78, u'Hyperbolic Sine',              4, u'sinh'),
        MathBlock(79, u'Hyperbolic Cosine',            4, u'cosh'),
        MathBlock(80, u'Inverse Hyperbolic Sine',      4, u'asinh'),
        MathBlock(81, u'Inverse Hyperbolic Cosine',    4, u'acosh'),
        MathBlock(82, u'Inverse Hyperbolic Tangent',   4, u'atanh'),
        ]

    return {x.name: x for x in blocks}

if not hasattr(MathBlockSet, '_all_blocks'):
    MathBlockSet._all_blocks = _generate_math_blocks_dict()



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter_value import ParameterValue

import json

class DataFileParameterValue(ParameterValue):
    """Combo box parameter description for analysis template

    :param Eureqa eureqa: A eureqa connection.
    :param str id: The id of the parameter.
    :param str file_path: The path to the file that should be uploaded to the server. If running locally, this path will be used to retreive the file directly.

    :var str id: The id of the parameter.
    :var str ~analysis_templates.DataFileParameterValue.file: the contents of the uploaded file
    :var str ~analysis_templates.DataFileParameterValue.value: the file id of the file on the server

    """

    def __init__(self, eureqa, id, file_path):
        """Initializes a new instance of the ~DataFileParameterValue class
        """
        self._eureqa = eureqa
        ParameterValue.__init__(self, id, None, "data_file")
        self.file_path=file_path
        self._execution = None

    def _to_json(self):
        body = {}
        ParameterValue._to_json(self, body)
        if self.file_path is not None:
            body['file_path'] = self.file_path
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @property
    def file(self):
        contents = ""
        if self._execution is None:
            assert self.file_path is not None
            f = open(self.file_path, 'rb')
            contents = f.read()
            f.close()
        else:
            contents = self._eureqa._session.execute('/analysis_templates/%s/executions/%s/files/%s' % (self._execution.template_id, self._execution._id, self.id), 'GET')
        return contents

    def _add_file_to_template(self, template):
        # This is used when calling 'execute()' directly on a template.
        # In that case, we have no matching DataFileParameter object to upload the file for us.
        # So we need a workaround to get the file data uploaded.
        f = open(self.file_path, 'rb')
        resp = self._eureqa._session.execute('/analysis_templates/%s/add_file'% (template._id), 'POST', files={'file': (self.file_path,f)})
        self.value = resp['file_id']

    @staticmethod
    def _from_json(eureqa, body, execution=None):
        param = DataFileParameterValue(eureqa, None, None)
        ParameterValue._from_json(param, body)
        param.file_path = body.get('file_path')
        if param._type != "data_file": raise Exception("Invalid type '%s' specified for data file parameter value" % param._type)
        param._execution = execution
        return param



#!/usr/bin/env python

# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This is a companion script for "install_analysis_template.py".
# Their implementations overlap substantially.

import os
import sys

## Implementation lives inside the API so that it can be imported generally
if __package__:
    # We're being run as "python -m eureqa run"
    from .analysis_templates.runner.client import main
else:
    # We're being run as "eureqa/run_analysis_template.py"
    sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
    from eureqa.analysis_templates.runner.client import main
    
## If we're being run, do stuff.
## (If we're being imported as a library, don't do stuff.)
if __name__ == "__main__":
    main(sys.argv)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component

class ModalLink(_Component):
    """Represents a link which opens a popup window (modal) which can contain other components.

    The use of this class is typically hidden by the Modal classes link method.

    Example:
    # create a modal which is activated if a link is clicked
    text_block = TextBlock("This text is inside the modal")
    modal = Modal(title="This is the modal title", size='medium', icon_file_path='path/to/icon', component=text_block)
    h = HtmlBlock('Click this link to open a modal: {0}'.format(analysis.html_ref(modal.link('link text'))))
    layout.add_component(h)

    # create another link to the above modal
    h = HtmlBlock('OR click this other link to open the same modal: {0}'.format(analysis.html_ref(modal.link('different link text'))))
    layout.add_component(h)

    :param eureqa.analysis.components.modal.Modal modal: the Modal which this component will link to
    :param str link_text: The HTML which the link to open the modal will be wrapped around
    """

    _component_type_str = 'MODAL_LINK'

    def __init__(self, modal=None, link_text=None,
                 _analysis=None, _component_id=None, _component_type=None):
        if modal is not None:
            self._modal_component = modal
        self._link_text = link_text or 'Modal Link'

        super(ModalLink, self).__init__(_analysis=_analysis, _component_id=_component_id,
                                        _component_type=_component_type)

        if hasattr(self, '_modal_component') and hasattr(self._modal_component, '_component_id'):
            self._modal_component_id = modal._component_id

    @property
    def link_text(self):
        """
        Text of the link as rendered by the component
        """
        return self._link_text

    @property
    def modal_component_id(self):
        """
        The ID of the modal_component to be linked to
        """
        return getattr(self, "_modal_component_id", None)

    def _walk_children(self):
        if hasattr(self, "_modal_component"):
            yield self._modal_component

    def _register(self, analysis):
        if hasattr(self, '_modal_component') and hasattr(self._modal_component, "_component_id"):
            self._modal_component_id = self._modal_component._component_id

        super(ModalLink, self)._register(analysis)

    def _associate_with_table(self, analysis):
        content_json = []
        if hasattr(self, '_modal_component'):
            content_json += self._modal_component._associate_with_table(analysis)
            self._modal_component_id = self._modal_component._component_id
        return super(ModalLink, self)._associate_with_table(analysis) + content_json

    def _fields(self):
        return super(ModalLink, self)._fields() + [ 'link_text', 'modal_component_id' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class ErrorMetrics(object):
    """
    Stores the value of the error of an expression as compared to a model
    by a variety of different types of error-evaluation methods.

    :param float mean_absolute_error: Mean Absolute Error
    :param float r2_goodness_of_fit: R^2 Goodness of Fit
    :param float correlation_coefficient: Correlation Coefficient
    :param float maximum_absolute_error: Maximum Absolute Error
    :param float signed_difference_between_lhs_and_rhs: Signed Difference Between LHS and RHS
    :param float area_under_roc_error: Area Under ROC Curve
    :param float log_loss_error: Log Loss Error
    :param float rank_correlation_1_minus_r: Rank Correlation
    :param float mean_square_error: Mean Squared Error
    :param float mean_squared_error_auc_hybrid: Mean Squared Error for Classification

    :var float mean_absolute_error: Mean Absolute Error
    :var float r2_goodness_of_fit: R^2 Goodness of Fit
    :var float correlation_coefficient: Correlation Coefficient
    :var float maximum_absolute_error: Maximum Absolute Error
    :var float signed_difference_between_lhs_and_rhs: Signed Difference Between LHS and RHS
    :var float area_under_roc_error: Area Under ROC Curve
    :var float log_loss_error: Log Loss Error
    :var float rank_correlation_1_minus_r: Rank Correlation
    :var float mean_square_error: Mean Squared Error
    :var float mean_squared_error_auc_hybrid: Mean Squared Error for Classification
    """

    def __init__(self,
                 mean_absolute_error = None,
                 r2_goodness_of_fit = None,
                 correlation_coefficient = None,
                 maximum_absolute_error = None,
                 signed_difference_between_lhs_and_rhs = None,
                 area_under_roc_error = None,
                 log_loss_error = None,
                 rank_correlation_1_minus_r = None,
                 mean_square_error = None,
                 mean_squared_error_auc_hybrid = None,
                 _body = None):
        """ ErrorMetrics init """

        self.mean_absolute_error = mean_absolute_error
        self.r2_goodness_of_fit = r2_goodness_of_fit
        self.correlation_coefficient = correlation_coefficient
        self.maximum_absolute_error = maximum_absolute_error
        self.signed_difference_between_lhs_and_rhs = signed_difference_between_lhs_and_rhs
        self.area_under_roc_error = area_under_roc_error
        self.log_loss_error = log_loss_error
        self.rank_correlation_1_minus_r = rank_correlation_1_minus_r
        self.mean_square_error = mean_square_error
        self.mean_squared_error_auc_hybrid = mean_squared_error_auc_hybrid
        self._series_id_value = None
        if _body is not None:
            self._from_json(_body)

    _metric_map = {
        'Mean Absolute Error': 'mean_absolute_error',
        'Mean Squared Error': 'mean_square_error',
        'R^2 Goodness of Fit': 'r2_goodness_of_fit',
        'Correlation Coefficient': 'correlation_coefficient',
        'Maximum Absolute Error': 'maximum_absolute_error',
        'Mean Logarithm Squared Error': '_mean_logarithm_squared_error',
        'Median Absolute Error': '_median_error',
        'Interquartile Mean Absolute Error': '_interquartile_mean_absolute_error',
        'Signed Difference Between LHS and RHS': 'signed_difference_between_lhs_and_rhs',
        'Implicit Derivative Error': '_implicit_derivative_error',
        'AIC Squared Error': '_squared_error_aic',
        'AIC Absolute Error': '_absolute_error_aic',
        'Area Under ROC Curve': 'area_under_roc_error',
        'Log Loss Error': 'log_loss_error',
        'Hinge Loss Error': '_hinge_loss_error',
        'Rank Correlation': 'rank_correlation_1_minus_r',
        'Slope Absolute Error': '_slope_absolute_error',
        'Mean Squared Error for Classification': 'mean_squared_error_auc_hybrid',
        'Maximum Classification Accuracy': '_maximum_classification_accuracy'
    }

    @property
    def series_id_value(self):
        return self._series_id_value

    def _from_json(self, json):
        self._body = json
        self._series_id_value = json.get('series_id_value')
        self._model = json.get('model')
        for elt in json['metrics']:
            setattr(self, self._metric_map[elt['metric_name']], elt['metric_value'])

    def _to_json(self):
        metrics = []
        for longname, slug in self._metric_map.iteritems():
            if hasattr(self, slug):
                metrics.append({'metric_name': longname, 'metric_value': getattr(self, slug)})
        return { 'metrics':         metrics,
                 'series_id_value': self.series_id_value}

"""Available error metrics:"""


def mean_absolute_error():
    """Mean Absolute Error error metric"""

    return 'Mean Absolute Error'

def r2_goodness_of_fit():
    """R^2 Goodness of Fit error metric"""

    return 'R^2 Goodness of Fit'

def correlation_coefficient():
    """Correlation Coefficient error metric"""

    return 'Correlation Coefficient'

def maximum_absolute_error():
    """Maximum Absolute Error error metric"""

    return 'Maximum Absolute Error'

def _mean_logarithm_squared_error():
    """Mean Logarithm Squared Error error metric"""

    return 'Mean Logarithm Squared Error'

def _median_error():
    """Median Absolute Error error metric"""

    return 'Median Absolute Error'

def _interquartile_mean_absolute_error():
    """Interquartile Mean Absolute Error error metric"""

    return 'Interquartile Mean Absolute Error'

def signed_difference_between_lhs_and_rhs():
    """Signed Difference Between LHS and RHS error metric"""

    return 'Signed Difference Between LHS and RHS'

def _implicit_derivative_error():
    """Implicit Derivative Error error metric"""

    return 'Implicit Derivative Error'

def _squared_error_aic():
    """AIC Squared Error error metric"""

    return 'AIC Squared Error'

def _absolute_error_aic():
    """AIC Absolute Error error metric"""

    return 'AIC Absolute Error'

def area_under_roc_error():
    """Area Under ROC Curve error metric"""

    return 'Area Under ROC Curve'

def log_loss_error():
    """Log Loss Error error metric"""

    return 'Log Loss Error'

def _hinge_loss_error():
    """Hinge Loss Error error metric"""

    return 'Hinge Loss Error'

def rank_correlation_1_minus_r():
    """Rank Correlation error metric"""

    return 'Rank Correlation'

def _slope_absolute_error():
    """Slope Absolute Error error metric"""

    return 'Slope Absolute Error'

def mean_square_error():
    """Mean Squared Error error metric"""

    return 'Mean Squared Error'

def mean_squared_error_auc_hybrid():
    """Mean Squared Error for Classification error metric"""

    return 'Mean Squared Error for Classification'

def _maximum_classification_accuracy():
    """Maximum Classification Accuracy error metric"""

    return 'Maximum Classification Accuracy'



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
from datetime import datetime

class VariableDetails:
    """Holds information about a variable. Obtained through the :py:meth:`~eureqa.data_source.DataSource.get_variable_details` method
    of a :py:class:`~eureqa.data_source.DataSource` . Use this class to find mathematic properties of a variable and to change a variable's
    name.

    :var str `~eureqa.variable.VariableDetails.name`: The name of the variable.
    :var str `~eureqa.variable.VariableDetails.expression`: The expression used to create the variable if it is a custom variable
    :var float/datetime `~eureqa.variable.VariableDetails.min_value`: The smallest value. Will be a datetime if datatype is 'datetime'.
    :var float/datetime `~eureqa.variable.VariableDetails.max_value`: The largest value. Will be a datetime if datatype is 'datetime'.
    :var float/datetime `~eureqa.variable.VariableDetails.mean_value`: The mean value. Will be a datetime if datatype is 'datetime'.
    :var float `~eureqa.variable.VariableDetails.standard_deviation`: The standard deviation of the values.
    :var int `~eureqa.variable.VariableDetails.distinct_values`: The number of distinct values.
    :var int `~eureqa.variable.VariableDetails.missing_values`: The number of missing values.
    :var int `~eureqa.variable.VariableDetails.total_rows`: The total rows occupied by the variable.
    :var str `~eureqa.variable.VariableDetails.datatype`: Whether it is a binary or numeric variable
    :var int `~eureqa.variable.VariableDetails.num_zeroes`: The number of zero values.
    :var int `~eureqa.variable.VariableDetails.num_ones`: The number of one values.
    :var `~eureqa.data_source.DataSource` `~eureqa.variable_details.VariableDetails.data_source`: The DataSource the VariableDetails belongs to
    """

    def __init__(self, body, data_source):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._id = body['variable_id']
        self.name = body['variable_name']
        self.expression = body.get('expression')
        self.standard_deviation = body['standard_deviation']
        self.distinct_values = body['distinct_values']
        self.missing_values = body['missing_values']
        self.num_zeroes = body['num_zeros']
        self.total_rows = body['total_rows']
        self.datatype = body['datatype']
        self._min_value = body['min_value']
        self._max_value = body['max_value']
        self._mean_value = body['mean_value']
        self.num_ones = body.get('num_ones')
        self.data_source = data_source
        self.derivation_type = body['type']
        self.seasonal_target_variable = body.get('seasonal_target_variable')
        self.seasonal_period = body.get('seasonal_period')
        self._body = body

    def _to_json(self):
        body = {
            'variable_id' : self._id,
            'variable_name' : self.name,
            'expression' : self.expression,
            'min_value' : self._min_value,
            'max_value' : self._max_value,
            'mean_value' : self._mean_value,
            'standard_deviation' : self.standard_deviation,
            'distinct_values' : self.distinct_values,
            'missing_values' : self.missing_values,
            'num_zeros' : self.num_zeroes,
            'total_rows' : self.total_rows,
            'datatype' : self.datatype,
            'num_ones' : self.num_ones,
            'datasource_id' : self.data_source._data_source_id,
            'type' : self.derivation_type,
            'seasonal_target_variable': self.seasonal_target_variable,
            'seasonal_period': self.seasonal_period
        }
        return body

    def __str__(self):
        body = self._to_json()
        body['min_value'] = self._get_formatted_value(self._min_value, True)
        body['max_value'] = self._get_formatted_value(self._max_value, True)
        body['mean_value'] = self._get_formatted_value(self._mean_value, True)
        return json.dumps(body, indent=4)

    def _get_formatted_value(self, val, dates_as_strings=False):
        if self.datatype == 'datetime':
            date = datetime.utcfromtimestamp(val/1000)
            if dates_as_strings:
                return str(date)
            return date
        return val

    @property
    def min_value(self):
        return self._get_formatted_value(self._min_value)

    @property
    def max_value(self):
        return self._get_formatted_value(self._max_value)

    @property
    def mean_value(self):
        return self._get_formatted_value(self._mean_value)


class SeasonalVariableDetails(VariableDetails):
    def __init__(self, body, data_source):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        VariableDetails.__init__(self, body, data_source)
        self.seasonal_target_variable = body['seasonal_target_variable']
        self.seasonal_period = body['seasonal_period']

    def _to_json(self):
        body = VariableDetails._to_json()

class Variable(VariableDetails):
    def __init__(*args, **kwargs):
        warnings.warn("Class 'Variable' has been renamed to 'VariableOptions'", DeprecationWarning)
        super(Variable, self).__init__(*args, **kwargs)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Internal module.  For use by the analysis_template command-line tools.
"""

import importlib
import warnings
import textwrap
import getpass
import json
import os

from eureqa.analysis_templates.parameters import Parameters

EUREQA_CONFIG_FILE = "eureqa_config.json"


DESCRIPTION = textwrap.dedent("""\
    This script is used to upload, run, and otherwise work locally with
    Analysis Templates.

    The first time this script is run, all argument values (other than the
    password) will be saved to a file named '%s' in the
    specified module directory.  Subsequent runs against a module will use the
    the options saved in that file by default so that you don't need to pass
    them in again.
    """ % EUREQA_CONFIG_FILE)

def add_module_args(parser):
    """ Arguments for describing the module to upload and run """

    module_args = parser.add_argument_group("Module Arguments")
    module_args.add_argument("module_dir",
                             help="Directory for module to upload.")

    # Field is named 'dependency_dir' (singular) even though it's a list
    # because the autogenerated help page describes it as a list of 'dependency_dir'
    # instances.  Makes the Python code slightly more confusing, though.
    module_args.add_argument("dependency_dir",
                             nargs='*',
                             help="Additional modules to bundle and install with the template.")

def add_server_args(parser):
    """ Arguments for connecting to the Eureqa server """

    server_args = parser.add_argument_group("Server Connection Arguments")
    server_args.add_argument("-u", "--username",
                             help="Username for Eureqa server")
    server_args.add_argument("-p", "--password",
                             help="Password for Eureqa server",
                             default='')
    server_args.add_argument("-k", "--api_key",
                             help="API Key for Eureqa server",
                             default='')
    server_args.add_argument("-o", "--organization",
                             help="Organization to upload the analysis template to (default: 'root')",
                             default=None)
    server_args.add_argument("--url",
                             help="URL to the Eureqa server (default: 'http://localhost:10002')",
                             default=None)
    server_args.add_argument("--verify_ssl_certificate",
                             choices=['true', 'false'],
                             help="Verify SSL certificate authenticity while connecting to server",
                             default='true')

## Args that are in either template_args or local_template_args
## Each may ignore saved parameters intended for the other
SHARED_ARGS = {"name", "description", "id", "parameter_values"}

def add_template_args(parser):
    """ Arguments for configuring the new template on the server """

    template_args = parser.add_argument_group("Template Configuration Arguments")
    template_args.add_argument("-n", "--name",
                               help="Name of the template to create (default: 'Example Template')",
                               default=None)
    template_args.add_argument("-d", "--description",
                               help="Description of the template to create (optional)",
                               default=None)
    template_args.add_argument("--id",
                               help="ID of template to update.  Optional; if omitted, will create a new template.",
                               default=None)
    template_args.add_argument("--icon",
                               help="Path to an icon to use for the analysis",
                               default=None)

def add_local_template_args(parser):
    """ Arguments to be passed to the new template when running the template locally """

    template_args = parser.add_argument_group("Template Configuration Arguments")
    template_args.add_argument("--parameter-values",
                               help="JSON string describing the values for the Parameters to be passed to the template at runtime",
                               type=json.loads,
                               dest="parameter_values",
                               default="[]")

def add_script_args(parser):
    """ Arguments that control details of the script's behavior """

    script_args = parser.add_argument_group("Script Usage Flags")
    script_args.add_argument("-q", "--quiet",
                             help="Only print the ID of the new analysis template to stdout.  Don't prompt for input on stdin.",
                             action='store_true')
    script_args.add_argument("--no-save",
                             help="Don't save arguments as %s" % os.path.join('[module_dir]', EUREQA_CONFIG_FILE),
                             dest="save", action='store_false')
    script_args.add_argument("--save-credentials",
                             help="Save login information to the Eureqa API credential store (not encrypted)",
                             dest="save_credentials", action='store_true')

def merge_defaults_from_json(module_dir, args):
    """ Read in default argument values from a user-created JSON config file """

    # 'hardcoded_defaults' are default values for fields that
    # want something fancier than None.
    # We can't use argparse for them because we want a semantic
    # difference between an argument's default value and
    # an argument that has been explicitly set to a value that
    # happens to equal its default.
    hardcoded_defaults = {
        'organization': 'root',
        'url': 'http://localhost:10002',
        'name': 'Example Template'
        }
    # Some defaults are legitimately None.  Import them.
    default_values = {k: v for k, v in args.__dict__.iteritems() if v is None}
    default_values.update(hardcoded_defaults)

    defaults_file = os.path.join(module_dir, EUREQA_CONFIG_FILE)

    cmdline_args = {k: v for k, v in args.__dict__.iteritems() if v is not None and v != []}

    if not os.path.isfile(defaults_file):
        json_conf = {}
    else:
        ## Load from JSON file
        try:
            with open(defaults_file) as f:
                json_conf = json.load(f)

                # Undo un-wrapping in write_defaults_to_json()
                if 'parameter_values' in json_conf:
                    json_conf['parameter_values'] = { 'parameters': json_conf['parameter_values'] }

        except Exception, e:
            warnings.warn("Error parsing configuration file: '%s'" % str(e))
            json_conf = {}  ## No config file for now

    argnames = set(args.__dict__.keys())

    ## Check for unrecognized stuff in the JSON
    unrecognized_keys = set(json_conf.keys()).difference(argnames).difference(SHARED_ARGS)
    if unrecognized_keys != set():
        warnings.warn("Unrecognized keys in %s: %s" % (EUREQA_CONFIG_FILE, ', '.join(unrecognized_keys)))

    conf = default_values
    conf.update(json_conf)
    conf.update(cmdline_args)

    return conf


def clean_and_populate_params(params, prompt):
    """ Make sure we have all requred params.  Prompt the user for missing values. """

    ## argparse should guarantee that this exists
    assert params['module_dir']
    assert params['organization']
    if not 'dependency_dir' in params:
        params['dependency_dir'] = []

    params['module_dir'] = os.path.abspath(params['module_dir'])
    params['dependency_dir'] = [os.path.abspath(d) for d in params['dependency_dir']]

    if not params.get('username', None):
        if not prompt:
            fail_no_prompt('--username')
        params['username'] = raw_input("Username: ")

    if not params.get('url', None):
        if not prompt:
            fail_no_prompt('--url')
        params['url'] = raw_input("RPC URL: ")

def fail_no_prompt(missing_arg):
    """ Error out and exit if we're missing an arg and can't prompt for it """

    raise RuntimeError, "Error:  Missing argument '%s'" % missing_arg

def module_name_from_module_dir(module_dir):
    return os.path.basename(module_dir)

def get_parameters_from_module(module_name):
    mod = importlib.import_module(module_name)
    return getattr(mod, 'template_parameters', Parameters)()

def write_defaults_to_json(params, module_dir):
    """ Write computed arguments to the Eureqa JSON file for this module """

    params = dict(params)  ## Make a copy, to avoid surprises in the caller

    ## Delete parameters that we don't want to persist

    # Saved through Eureqa's credentials store instead
    if 'password' in params:
        del params['password']

    # One-time command-line behavior
    if 'quiet' in params:
        del params['quiet']
    if 'save' in params:
        del params['save']
    if 'save_credentials' in params:
        del params['save_credentials']
    if 'verify_ssl_certificate' in params:
        del params['verify_ssl_certificate']

    # Redundant (we're writing to a file located at this path)
    if 'module_dir' in params:
        del params['module_dir']

    if 'parameter_values' in params:
        ## Un-wrap 'parameter_values' serialization format for easier-to-read storage
        try:
            params['parameter_values'] = params['parameter_values']['parameters']
        except:
            pass  ## Don't know what you're up to but apparently the server liked it?

    try:
        with open(os.path.join(module_dir, EUREQA_CONFIG_FILE), "w") as f:
            json.dump(params, f, indent=4, sort_keys=True)
    except Exception, e:
        warnings.warn("Unable to save settings to '%s': '%s'" % (EUREQA_CONFIG_FILE, str(e)))
